module Arxiv
  where

import Control.Applicative ((<$>))
import Control.Monad.Trans (liftIO)
import Control.Monad.Trans.Resource (MonadResource)
import Data.Conduit ((.|))
import Data.Either (Either(..))
import Data.Function ((&))
import Data.List (intercalate)
import Debug.Trace (trace)
import Network.Api.Arxiv (Expression(..),  Field(..), (/*/), (/+/),(/-/))
import Network.HTTP.Conduit (parseRequest)
import Network.HTTP.Simple as HT
import Network.HTTP.Types.Status
import Text.HTML.TagSoup
import Text.Read (readMaybe)
import qualified Conduit as C
import qualified Data.ByteString as B hiding (unpack)
import qualified Data.ByteString.Char8 as B  (unpack)
import qualified Data.Conduit as C
import qualified Data.Conduit.List as CL
import qualified Network.Api.Arxiv as Ax

------------------------------------------------------------
type Limit = Int
type Soup  = Tag String

batchSize :: Int
batchSize = 400

------------------------------------------------------------

-- apiTotal_simple :: [String] -> IO Int
--get :: MonadResource m => Maybe Limit -> [String] -> C.ConduitT () Result m ()
--get Nothing query = searchAxv (simpleQuery query)
--get (Just l) query = searchAxv (simpleQuery query) .| CL.take l

------------------------------------------------------------
-- | Simple Query search for terms in All fields
apiSimple :: Maybe Limit -> [String] -> IO [Result]
apiSimple limit query = execQuery limit $ simpleQuery query

apiSimpleC :: (Monad m, C.MonadIO m) => Maybe Limit -> [String] -> IO (Int, C.ConduitT () Result m ())
apiSimpleC Nothing query = searchAxv' $ simpleQuery query
apiSimpleC (Just limit) query = do
  (cnt, resC) <- searchAxv' $ simpleQuery query
  pure (cnt, resC .| C.takeC limit)

-- | Complex Query search for terms in specific fields to be defined by Ax.Query
apiComplex :: Maybe Limit -> Ax.Query -> IO [Result]
apiComplex limit query = execQuery limit query

------------------------------------------------------------
simpleQuery :: [String] -> Ax.Query
simpleQuery xs = Ax.Query { Ax.qExp = Just $ Exp $ Abs xs
                          , Ax.qIds = []
                          , Ax.qStart = 0
                          , Ax.qItems = batchSize }

complexQuery :: Ax.Query
complexQuery =
  let au = Exp $ Au ["Aaronson"]
      t1 = Exp $ Ti ["quantum"]
      t2 = Exp $ Ti ["complexity"]
      x  = au /*/ (t1 /+/ t2)
   in Ax.Query { Ax.qExp   = Just $ Exp $ All ["nuclear fusion", "plasma"]
               , Ax.qIds   = []
               , Ax.qStart = 0
               , Ax.qItems = 50
               }


execQuery :: Maybe Limit -> Ax.Query -> IO [Result]
execQuery Nothing  q = C.runConduitRes (searchAxv q .| CL.consume)
execQuery (Just l) q = C.runConduitRes (searchAxv q .| CL.take l)
-- execQuery q = C.runConduitRes (searchAxv' q .| outSnk)

----------------------------------------------------------------------
-- Execute query and start a source
----------------------------------------------------------------------
searchAxvBody ::  Ax.Query -> IO B.ByteString
searchAxvBody q =
  let s = Ax.mkQuery q
   in do rsp <- HT.httpBS =<< liftIO (parseRequest s)
         case getResponseStatus rsp of
           (Status 200 _) -> pure $ getResponseBody rsp
           st             -> error $ "Error:" ++ show st

searchAxv' :: (Monad m, C.MonadIO m) => Ax.Query -> IO (Int, C.ConduitT () Result m ())
searchAxv' q =
  let s = Ax.mkQuery q
   in do rsp <- HT.httpBS =<< liftIO (parseRequest s)
         case getResponseStatus rsp of
           (Status 200 _) -> case getSoup (getResponseBody rsp) of
             []          -> pure (0, C.yieldMany [])
             soups@(s:_) -> pure (Ax.totalResults soups, results q soups)
           st             -> error $ "Error:" ++ show st

searchAxv :: (Monad m, C.MonadIO m) => Ax.Query -> C.ConduitT () Result m ()
searchAxv q =
  let s = Ax.mkQuery q
   in do rsp <- HT.httpBS =<< liftIO (parseRequest s)
         case getResponseStatus rsp of
           (Status 200 _) -> case getSoup (getResponseBody rsp) of
             []          -> C.yieldMany []
             soups@(s:_) -> results q soups
           st             -> error $ "Error:" ++ show st

----------------------------------------------------------------------
-- Consume page by page
----------------------------------------------------------------------
getSoup :: B.ByteString -> [Soup]
getSoup b = parseTags $ B.unpack b


--getSoupC :: MonadResource m =>
--            B.ByteString -> C.ConduitT () a m [Soup]
getSoupC :: (Monad m, C.MonadIO m) => B.ByteString -> C.ConduitT () a m [Soup]
getSoupC b = concat <$> (C.yield b .| C.mapM_C (liftIO . putStrLn . show) .| toSoup .| CL.consume)


----------------------------------------------------------------------
-- Receive a ByteString and yield Soup
----------------------------------------------------------------------
--toSoup :: MonadResource m => C.ConduitT B.ByteString [Soup] m ()
toSoup :: Monad m => C.ConduitT B.ByteString [Soup] m ()
toSoup = C.awaitForever (C.yield . parseTags . B.unpack)

----------------------------------------------------------------------
-- Yield all entries and fetch next page
----------------------------------------------------------------------
results :: (Monad m, C.MonadIO m) =>
           Ax.Query -> [Soup] -> C.ConduitT () Result m ()
results q sp = do
  Ax.forEachEntryM sp (C.yield . mkResult)
  searchAxv (Ax.nextPage q)

----------------------------------------------------------------------
-- Get data and format
----------------------------------------------------------------------
data Result = Result { abstract         :: String
                     , authors          :: [Ax.Author]
                     , categories       :: [Ax.Category]
                     , doi              :: String
                     , id               :: String
                     , journal          :: String
                     , primaryCategory  :: Maybe Ax.Category
                     , publication_date :: String
                     , title            :: String
                     , total            :: Int
                     , url              :: String
                     , year             :: Maybe Int
                     } deriving (Show)

mkResult :: [Soup] -> Result
mkResult sp = let abstract         = Ax.getSummary         sp & clean'
                  authors          = Ax.getAuthors         sp
                  categories       = Ax.getCategories      sp
                  doi              = Ax.getDoi             sp
                  id               = Ax.getId              sp
                  journal          = Ax.getJournal         sp
                  primaryCategory  = Ax.getPrimaryCategory sp
                  publication_date = Ax.getPublished       sp
                  title            = Ax.getTitle           sp & clean'
                  total            = Ax.totalResults       sp
                  url              = Ax.getPdf             sp
                  year             = readMaybe $ Ax.getYear sp
               in ( Result { .. } )
  where clean' x = let x' = clean ['\n', '\r', '\t'] x
                    in if null x' then "Not found" else x'
        clean _ [] = []
        clean d (c:cs) | c `elem` d =  " " <> clean d cs
                       | otherwise  = c:clean d cs


----------------------------------------------------------------------
-- Sink results To print
----------------------------------------------------------------------
outSnk :: (Monad m, C.MonadIO m) => C.ConduitT Result C.Void m ()
outSnk = C.awaitForever (liftIO . putStrLn . show)
